@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # =============================================================================
-#     filter_functions
+#     qopt
 #     Copyright (C) 2020 Julian Teske, Forschungszentrum Juelich
 #
 #     This program is free software: you can redistribute it and/or modify
@@ -19,8 +19,13 @@
 #     Contact email: j.teske@fz-juelich.de
 # =============================================================================
 """
-This file contains classes and helper functions for the generation of noise
-traces.
+This file contains classes and helper functions for the sampling of noise
+distributions.
+
+The `NoiseTraceGenerator` class and its children are used by the `Solver` class
+to generate noise samples. They use various helper functions for the respective
+sampling methods. In the current version, the sampling of a quasi static
+distribution and the spectral noise density of fast noise is supported.
 
 Classes
 -------
@@ -113,7 +118,7 @@ def sample_1dim_gaussian_distribution(std1: float,
 
     Returns
     -------
-    selected_x: np.ndarray
+    selected_x: list of float
         Noise samples.
 
     """
@@ -286,8 +291,8 @@ class NoiseTraceGenerator(ABC):
     n_noise_operators: int, optional
         Number of noise operators. Default is 1.
 
-    noise_samples: None or np.ndarray, optional
-                   shape: (n_noise_operators, n_traces, n_samples_per_trace)
+    noise_samples: None or np.ndarray, optional,
+        shape (n_noise_operators, n_traces, n_samples_per_trace)
         Precalculated noise samples. Defaults to None.
 
     always_redraw_samples: bool
@@ -296,28 +301,10 @@ class NoiseTraceGenerator(ABC):
 
     Attributes
     ----------
-    noise_samples: np.ndarray,
-                   shape: (n_noise_operators, n_traces, n_samples_per_trace)
-        The noise samples stored or generated by this class.
-
-    n_samples_per_trace: int
-        Number of noise samples per trace.
-
-    n_traces: int
-        Number of noise traces.
-
-    n_noise_operators: int
-        Number of noise operators.
-
     always_redraw_samples: bool
         If true. The samples are always redrawn upon request. The stored samples
         are not returned.
 
-    Methods
-    -------
-    _sample_noise: None
-        Draw noise samples.
-
     """
 
     def __init__(self, n_samples_per_trace: int, always_redraw_samples: bool,
@@ -377,7 +364,7 @@ class NTGQuasiStatic(NoiseTraceGenerator):
 
     Parameters
     ----------
-    standard_deviation: List[float], len: (n_noise_operators)
+    standard_deviation: List[float], len (n_noise_operators)
         Standard deviations of the noise assumed on the noise operators.
 
     n_samples_per_trace: int
@@ -387,7 +374,7 @@ class NTGQuasiStatic(NoiseTraceGenerator):
         Number of noise traces. Default is 1.
 
     noise_samples: None or np.ndarray, optional
-                   shape: (n_noise_operators, n_traces, n_samples_per_trace)
+        shape (n_noise_operators, n_traces, n_samples_per_trace)
         Precalculated noise samples. Defaults to None.
 
     sampling_mode: {'uncorrelated_deterministic', 'monte_carlo'}, optional
@@ -400,7 +387,7 @@ class NTGQuasiStatic(NoiseTraceGenerator):
 
     Attributes
     ----------
-    standard_deviation: List[float], len: (n_noise_operators)
+    standard_deviation: List[float], len (n_noise_operators)
         Standard deviations of the noise assumed on the noise operators.
 
     See Also
@@ -493,8 +480,8 @@ class NTGColoredNoise(NoiseTraceGenerator):
         Number of noise operators. Default is 1.
 
     always_redraw_samples: bool
-        If true. The samples are always redrawn upon request. The stored samples
-        are not returned.
+        If true. The samples are always redrawn upon request. The stored
+        samples are not returned.
 
     noise_spectral_density: function
         The one-sided noise spectral density as function of frequency.
@@ -502,6 +489,11 @@ class NTGColoredNoise(NoiseTraceGenerator):
     dt: float
         Time distance between two adjacent samples.
 
+    low_frequency_extension_ratio: int, optional
+        When creating the time samples, the total time is multiplied with this
+        factor. This allows taking frequencies into account which are lower
+        than 1 / total time. Defaults to 1.
+
     Attributes
     ----------
     noise_spectral_density: function
@@ -527,7 +519,8 @@ class NTGColoredNoise(NoiseTraceGenerator):
                  dt: float,
                  n_traces: int = 1,
                  n_noise_operators: int = 1,
-                 always_redraw_samples: bool = True):
+                 always_redraw_samples: bool = True,
+                 low_frequency_extension_ratio: int = 1):
         super().__init__(n_traces=n_traces,
                          n_samples_per_trace=n_samples_per_trace,
                          noise_samples=None,
@@ -535,6 +528,10 @@ class NTGColoredNoise(NoiseTraceGenerator):
                          always_redraw_samples=always_redraw_samples)
         self.noise_spectral_density = noise_spectral_density
         self.dt = dt
+        if low_frequency_extension_ratio < 1:
+            raise ValueError("The low frequency extension ratio must be "
+                             "greater or equal to 1.")
+        self.low_frequency_extension_ratio = low_frequency_extension_ratio
         if hasattr(dt, "__len__"):
             raise ValueError('dt is supposed to be a scalar value!')
 
@@ -549,11 +546,12 @@ class NTGColoredNoise(NoiseTraceGenerator):
                              'trace!')
         noise_samples = fast_colored_noise(
             spectral_density=self.noise_spectral_density,
-            n_samples=self.n_samples_per_trace,
+            n_samples=
+            self.n_samples_per_trace * self.low_frequency_extension_ratio,
             output_shape=(self.n_noise_operators, self.n_traces),
             r_power_of_two=False,
             dt=self.dt)
-        self._noise_samples = noise_samples
+        self._noise_samples = noise_samples[:, :, :self.n_samples_per_trace]
 
     def plot_periodogram(self, n_average: int, scaling: str = 'density',
                          log_plot: Optional[str] = None, draw_plot=True):
