@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # =============================================================================
-#     filter_functions
+#     qopt
 #     Copyright (C) 2020 Julian Teske, Forschungszentrum Juelich
 #
 #     This program is free software: you can redistribute it and/or modify
@@ -19,14 +19,18 @@
 #     Contact email: j.teske@fz-juelich.de
 # =============================================================================
 """
-Optimization algorithms.
+This module contains interfaces to optimization algorithms or the algorithms
+themselves.
 
-This module implements the optimization algorithms for the optimal control
-problem.
+Currently it supports interfacing to least squares and minimization algorithms
+of the scipy package. Conditionally, a simulated annealing is supported, if the
+simanneal package is included in the environment. (See installation
+instructions on https://github.com/qutech/qopt) The classes for simulated
+annealing are in an experimental state and not well tested!
 
-Currently supported are:
+The `Optimizer` class uses the `Simulator` as interface to the cost functions
+evaluated after the simulation.
 
-    LS-TRF - Least squares, Trust Region Reflective
 
 Classes
 -------
@@ -36,9 +40,23 @@ Classes
 :class:`LeastSquaresOptimizer`
     An interface to scipy's least squares optimizer.
 
+:class:`ScalarMinimizingOptimizer`
+    An interface to scipy's minimize functions.
+
 :class:`WallTimeExceeded`
     Exception for exceeding the optimization's time limit.
 
+:class:`PulseAnnealer`
+    Helper class for `SimulatedAnnealing`.
+
+:class:`SimulatedAnnealing`
+    Simulated annealing as optimization. Experimental implementation. Not well
+    tested!
+
+:class:`SimulatedAnnealingScipy`
+    Simulated annealing based on scipy functions. Experimental implementation.
+    Not well tested!
+
 Notes
 -----
 The implementation was inspired by the optimal control package of QuTiP [1]_
@@ -57,10 +75,20 @@ import scipy
 import scipy.optimize
 import time
 from abc import ABC, abstractmethod
-from typing import Dict, Optional, Callable, List, Union
+from typing import Dict, Optional, Callable, List, Union, Sequence
+from unittest import mock
+from warnings import warn
+
+from qopt import optimization_data, simulator, performance_statistics
+
+
+try:
+    import simanneal
+except ImportError:
+    warn('simanneal not installed. '
+         'SimulatedAnnealing is not available')
+    simanneal = mock.Mock()
 
-from qopt import optimization_data, simulator
-import simanneal
 
 default_termination_conditions = {
     "min_gradient_norm": 1e-7,
@@ -80,6 +108,31 @@ class WallTimeExceeded(Exception):
 class Optimizer(ABC):
     """ Abstract base class for the optimizer.
 
+    Parameters
+    ----------
+    system_simulator : Simulator
+        The simulator is the interface to the simulation.
+
+    termination_cond: dict, optional
+        The termination conditions of the optimization.
+
+    save_intermediary_steps: bool, optional
+        If True, then the results from intermediary steps are stored. Defaults
+        to True.
+
+    cost_func_weights: list of float, optional
+        The cost functions are multiplied with these weights during the
+        optimisation.
+
+    use_jacobian_function: bool, optional
+        If set to true, then the jacobians are calculated analytically.
+        Defaults to True.
+
+    store_optimizer: bool, optional
+        If True, then the optimizer stores itself in the result class.
+        Defaults to False
+
+
     Attributes
     ----------
     system_simulator : Simulator
@@ -89,20 +142,33 @@ class Optimizer(ABC):
         The shape of the control amplitudes is saved and used for the
         cost functions while the optimization function might need them flatted.
 
-    TODO:
-        * implement termination conditions such as wall time!
+    cost_func_weights: list of float, optional
+        The cost functions are multiplied with these weights during the
+        optimisation.
+
+    use_jacobian_function: bool, optional
+        If set to true, then the jacobians are calculated analytically.
+
+    store_optimizer: bool, optional
+        If True, then the optimizer stores itself in the result class.
+        Defaults to False
+
     """
 
     def __init__(
             self,
             system_simulator: Optional[simulator.Simulator] = None,
             termination_cond: Optional[Dict] = None,
-            save_intermediary_steps: bool = False):
+            save_intermediary_steps: bool = True,
+            cost_func_weights: Optional[Sequence[float]] = None,
+            use_jacobian_function=True,
+            store_optimizer: bool = False
+    ):
         self.system_simulator = system_simulator
-        if termination_cond is None:
-            self.termination_conditions = default_termination_conditions
-        else:
-            self.termination_conditions = termination_cond
+        self.use_jacobian_function = use_jacobian_function
+        self.termination_conditions = default_termination_conditions
+        if termination_cond is not None:
+            self.termination_conditions.update(**termination_cond)
 
         self.optim_iter_summary = None
         self.pulse_shape = ()
@@ -115,8 +181,17 @@ class Optimizer(ABC):
 
         # flags:
         self.save_intermediary_steps = save_intermediary_steps
+        self.store_optimizer = store_optimizer
 
-    def cost_fktn_wrapper(self, optimization_parameters):
+        self.cost_func_weights = cost_func_weights
+
+        if self.cost_func_weights is not None:
+            self.cost_func_weights = np.asarray(
+                self.cost_func_weights).flatten()
+            if len(self.cost_func_weights) == 0:
+                self.cost_func_weights = None
+
+    def cost_func_wrapper(self, optimization_parameters):
         """Wraps the cost function given by the simulator class.
 
         The relevant information for the analysis is saved.
@@ -128,7 +203,7 @@ class Optimizer(ABC):
 
         Returns
         -------
-        costs: np.array
+        costs: np.array, shape (n_fun)
             Cost values.
 
         """
@@ -150,6 +225,10 @@ class Optimizer(ABC):
             self._min_costs_par = optimization_parameters.reshape(
                 self.pulse_shape[::-1]).T
 
+        # apply the cost function weights after saving the values.
+        if self.cost_func_weights is not None:
+            costs *= self.cost_func_weights
+
         self._n_cost_fkt_eval += 1
         return costs
 
@@ -165,7 +244,7 @@ class Optimizer(ABC):
 
         Returns
         -------
-        jacobian: np.array
+        jacobian: np.array, shape (num_func, num_t * num_amp)
             Jacobian of the cost functions.
 
         """
@@ -180,11 +259,17 @@ class Optimizer(ABC):
         jacobian = jacobian.reshape(
             (jacobian.shape[0], jacobian.shape[1] * jacobian.shape[2]))
 
+        # apply the cost function weights after saving the values.
+        if self.cost_func_weights is not None:
+            jacobian = np.einsum('ab, a -> ab', jacobian,
+                                 self.cost_func_weights)
+
         self._n_jac_fkt_eval += 1
         return jacobian
 
     @abstractmethod
-    def run_optimization(self, initial_control_amplitudes: np.ndarray) \
+    def run_optimization(self, initial_control_amplitudes: np.ndarray,
+                         verbose) \
             -> optimization_data.OptimizationResult:
         """Runs the optimization of the control amplitudes.
 
@@ -192,6 +277,8 @@ class Optimizer(ABC):
         ----------
         initial_control_amplitudes : array
             shape (num_t, num_ctrl)
+        verbose
+            Verbosity of the run. Depends on which optimizer is used.
 
         Returns
         -------
@@ -224,11 +311,55 @@ class Optimizer(ABC):
                 )
         self._opt_start_time = time.time()
         if self.system_simulator.stats is not None:
+            # If the system simulator wants to write down statistics, then
+            # initialise a fresh instance
+            self.system_simulator.stats = \
+                performance_statistics.PerformanceStatistics()
             self.system_simulator.stats.start_t_opt = float(
                 self._opt_start_time)
             self.system_simulator.stats.indices = \
                 self.system_simulator.cost_indices
 
+    def write_state_to_result(self):
+        """ Writes the current state into an instance of 'OptimizationResult'.
+
+        Intended for saving progress when terminating the optimization in an
+        unexpected way.
+
+        Returns
+        -------
+        result: optimization_data.OptimizationResult
+            The current result of the optimization.
+
+        """
+        if self.system_simulator.stats is not None:
+            self.system_simulator.stats.end_t_opt = time.time()
+
+        if self.use_jacobian_function:
+            jac_norm = np.linalg.norm(
+                self.cost_jacobian_wrapper(self._min_costs_par))
+        else:
+            jac_norm = 0
+
+        if self.store_optimizer:
+            storage_opt = self
+        else:
+            storage_opt = None
+
+        optim_result = optimization_data.OptimizationResult(
+            final_cost=self._min_costs,
+            indices=self.system_simulator.cost_indices,
+            final_parameters=self._min_costs_par,
+            final_grad_norm=jac_norm,
+            num_iter=self._n_cost_fkt_eval,
+            termination_reason='Maximum Wall Time Exceeded',
+            status=5,
+            optimizer=storage_opt,
+            optim_summary=self.optim_iter_summary,
+            optimization_stats=self.system_simulator.stats
+        )
+        return optim_result
+
 
 class LeastSquaresOptimizer(Optimizer):
     """
@@ -249,17 +380,13 @@ class LeastSquaresOptimizer(Optimizer):
         The optimization method used. Currently implemented are:
         - 'trf': A trust region optimization algorithm. This is the default.
 
-    bounds: array or list of boundaries, optional
+    bounds: list of array-like, optional
+        Attention: The boundary format can vary between optimizers!
         The boundary conditions for the pulse optimizations. If none are given
-        then the pulse is assumed to take any real value.
-
-    use_jacobian_function: bool, optional
-        If set to true, then the jacobians are calculated analytically.
-
-
-    TODO:
-        * in the handling of a wall time exceeded message, the last parameters
-            * are stored instead of the best ones.
+        then the pulse is assumed to take any real value. The boundaries are
+        given as a list of two arrays. The first array specifies the upper
+        and the second array specifies the lower bounds. Single parameters
+        can be excepted by using np.inf with appropriate sign.
 
     """
 
@@ -267,51 +394,54 @@ class LeastSquaresOptimizer(Optimizer):
             self,
             system_simulator: Optional[simulator.Simulator] = None,
             termination_cond: Optional[Dict] = None,
-            save_intermediary_steps: bool = False,
+            save_intermediary_steps: bool = True,
             method: str = 'trf',
             bounds: Union[np.ndarray, List, None] = None,
-            use_jacobian_function=True):
+            use_jacobian_function=True,
+            cost_func_weights: Optional[Sequence[float]] = None,
+            store_optimizer: bool = False):
         super().__init__(system_simulator=system_simulator,
                          termination_cond=termination_cond,
-                         save_intermediary_steps=save_intermediary_steps)
+                         save_intermediary_steps=save_intermediary_steps,
+                         cost_func_weights=cost_func_weights,
+                         use_jacobian_function=use_jacobian_function,
+                         store_optimizer=store_optimizer)
         self.method = method
         self.bounds = bounds
-        self.use_jacobian_function = use_jacobian_function
 
-    def run_optimization(self, initial_control_amplitudes: np.array) \
-            -> optimization_data.OptimizationResult:
+    def run_optimization(self, initial_control_amplitudes: np.array,
+                         verbose: int = 0) -> optimization_data.OptimizationResult:
         """See base class. """
         super().prepare_optimization(
             initial_optimization_parameters=initial_control_amplitudes)
 
+        if self.use_jacobian_function:
+            jac = super().cost_jacobian_wrapper
+        else:
+            jac = '2-point'
+
         try:
-            if self.use_jacobian_function:
-                result = scipy.optimize.least_squares(
-                    fun=super().cost_fktn_wrapper,
-                    x0=initial_control_amplitudes.T.flatten(),
-                    jac=super().cost_jacobian_wrapper,
-                    bounds=self.bounds,
-                    method=self.method,
-                    ftol=self.termination_conditions["min_cost_gain"],
-                    xtol=self.termination_conditions["min_amplitude_change"],
-                    gtol=self.termination_conditions["min_gradient_norm"],
-                    max_nfev=self.termination_conditions["max_iterations"]
-                )
-            else:
-                result = scipy.optimize.least_squares(
-                    fun=super().cost_fktn_wrapper,
-                    x0=initial_control_amplitudes.T.flatten(),
-                    bounds=self.bounds,
-                    method=self.method,
-                    ftol=self.termination_conditions["min_cost_gain"],
-                    xtol=self.termination_conditions["min_amplitude_change"],
-                    gtol=self.termination_conditions["min_gradient_norm"],
-                    max_nfev=self.termination_conditions["max_iterations"]
-                )
+            result = scipy.optimize.least_squares(
+                fun=super().cost_func_wrapper,
+                x0=initial_control_amplitudes.T.flatten(),
+                jac=jac,
+                bounds=self.bounds,
+                method=self.method,
+                ftol=self.termination_conditions["min_cost_gain"],
+                xtol=self.termination_conditions["min_amplitude_change"],
+                gtol=self.termination_conditions["min_gradient_norm"],
+                max_nfev=self.termination_conditions["max_iterations"],
+                verbose=verbose
+            )
 
             if self.system_simulator.stats is not None:
                 self.system_simulator.stats.end_t_opt = time.time()
 
+            if self.store_optimizer:
+                storage_opt = self
+            else:
+                storage_opt = None
+
             optim_result = optimization_data.OptimizationResult(
                 final_cost=result.fun,
                 indices=self.system_simulator.cost_indices,
@@ -321,32 +451,185 @@ class LeastSquaresOptimizer(Optimizer):
                 num_iter=result.nfev,
                 termination_reason=result.message,
                 status=result.status,
-                optimizer=self,
+                optimizer=storage_opt,
                 optim_summary=self.optim_iter_summary,
                 optimization_stats=self.system_simulator.stats
             )
         except WallTimeExceeded:
-            if self.system_simulator.stats is not None:
-                self.system_simulator.stats.end_t_opt = time.time()
+            optim_result = self.write_state_to_result()
 
-            if self.use_jacobian_function:
-                jac_norm = np.linalg.norm(
-                    self.cost_jacobian_wrapper(self._min_costs_par))
-            else:
-                jac_norm = 0
+        return optim_result
 
-            optim_result = optimization_data.OptimizationResult(
-                final_cost=self._min_costs,
-                indices=self.system_simulator.cost_indices,
-                final_parameters=self._min_costs_par,
-                final_grad_norm=jac_norm,
-                num_iter=self._n_cost_fkt_eval,
-                termination_reason='Maximum Wall Time Exceeded',
-                status=5,
-                optimizer=self,
-                optim_summary=self.optim_iter_summary,
-                optimization_stats=self.system_simulator.stats
-            )
+
+class ScalarMinimizingOptimizer(Optimizer):
+    """ Interfaces to the minimize functions of the optimization package in
+    scipy.
+
+    Parameters
+    ----------
+    method: string
+        Takes methods implemented by scipy.optimize.minimize.
+
+    bounds: sequence, optional
+        Attention: The boundary format can vary between optimizers!
+        The boundary conditions for the pulse optimizations. If none are given
+        then the pulse is assumed to take any real value. The boundaries are
+        given as a sequence of (min, max) pairs for each element. Defaults to
+        None.
+
+    """
+    def __init__(
+            self,
+            system_simulator: Optional[simulator.Simulator] = None,
+            termination_cond: Optional[Dict] = None,
+            save_intermediary_steps: bool = True,
+            method: str = 'L-BFGS-B',
+            bounds: Union[np.ndarray, List, None] = None,
+            use_jacobian_function=True,
+            cost_func_weights: Optional[Sequence[float]] = None,
+            store_optimizer: bool = False
+    ):
+        super().__init__(system_simulator=system_simulator,
+                         termination_cond=termination_cond,
+                         save_intermediary_steps=save_intermediary_steps,
+                         cost_func_weights=cost_func_weights,
+                         use_jacobian_function=use_jacobian_function,
+                         store_optimizer=store_optimizer)
+        self.method = method
+        self.bounds = bounds
+
+    def cost_func_wrapper(self, optimization_parameters):
+        """ Evalutes the cost function.
+
+         The total cost function is defined as the sum of cost functions.
+
+         """
+        costs = super().cost_func_wrapper(optimization_parameters)
+        scalar_costs = np.sum(costs)
+        return scalar_costs
+
+    def cost_jacobian_wrapper(self, optimization_parameters):
+        """ The Jacobian reduced to the gradient.
+
+        The gradient is calculated by summation over the Jacobian along the
+        function axis, because the total cost function is defined as the sum
+        of cost functions.
+
+        Returns
+        -------
+        gradient: numpy array, shape (num_t * num_amp)
+            The gradient of the costs in the 2 norm.
+
+        """
+        jac = super().cost_jacobian_wrapper(optimization_parameters)
+        grad = (np.sum(jac, axis=0))
+        return grad
+
+    def run_optimization(self, initial_control_amplitudes: np.array,
+                         verbose: bool = False) -> optimization_data.OptimizationResult:
+        super().prepare_optimization(
+            initial_optimization_parameters=initial_control_amplitudes)
+
+        if self.use_jacobian_function:
+            jac = self.cost_jacobian_wrapper
+        else:
+            jac = None
+
+        if self.method == 'L-BFGS-B':
+            try:
+                result = scipy.optimize.minimize(
+                    fun=self.cost_func_wrapper,
+                    x0=initial_control_amplitudes.T.flatten(),
+                    jac=jac,
+                    bounds=self.bounds,
+                    method=self.method,
+                    options={
+                        'ftol': self.termination_conditions["min_cost_gain"],
+                        'gtol': self.termination_conditions["min_gradient_norm"],
+                        'maxiter': self.termination_conditions["max_iterations"],
+                        'disp': verbose
+                    }
+                )
+
+                if self.store_optimizer:
+                    storage_opt = self
+                else:
+                    storage_opt = None
+
+                optim_result = optimization_data.OptimizationResult(
+                    final_cost=result.fun,
+                    indices=self.system_simulator.cost_indices,
+                    final_parameters=result.x.reshape(
+                        self.pulse_shape[::-1]).T,
+                    final_grad_norm=np.linalg.norm(result.jac),
+                    num_iter=result.nfev,
+                    termination_reason=result.message,
+                    status=result.status,
+                    optimizer=storage_opt,
+                    optim_summary=self.optim_iter_summary,
+                    optimization_stats=self.system_simulator.stats
+                )
+            except WallTimeExceeded:
+                optim_result = self.write_state_to_result()
+
+        elif self.method == 'Nelder-Mead':
+            try:
+                result = scipy.optimize.minimize(
+                    fun=self.cost_func_wrapper,
+                    x0=initial_control_amplitudes.T.flatten(),
+                    bounds=self.bounds,
+                    method=self.method,
+                    options={
+                        'maxiter': self.termination_conditions[
+                            "max_iterations"]},
+                )
+
+                if self.store_optimizer:
+                    storage_opt = self
+                else:
+                    storage_opt = None
+
+                optim_result = optimization_data.OptimizationResult(
+                    final_cost=result.fun,
+                    indices=self.system_simulator.cost_indices,
+                    final_parameters=result.x.reshape(
+                        self.pulse_shape[::-1]).T,
+                    num_iter=result.nfev,
+                    termination_reason=result.message,
+                    status=result.status,
+                    optimizer=storage_opt,
+                    optim_summary=self.optim_iter_summary,
+                    optimization_stats=self.system_simulator.stats
+                )
+            except WallTimeExceeded:
+                optim_result = self.write_state_to_result()
+
+        else:
+            try:
+                result = scipy.optimize.minimize(
+                    fun=self.cost_func_wrapper,
+                    x0=initial_control_amplitudes.T.flatten(),
+                    bounds=self.bounds,
+                    method=self.method
+                )
+
+                optim_result = optimization_data.OptimizationResult(
+                    final_cost=result.fun,
+                    indices=self.system_simulator.cost_indices,
+                    final_parameters=result.x.reshape(
+                        self.pulse_shape[::-1]).T,
+                    num_iter=result.nfev,
+                    termination_reason=result.message,
+                    status=result.status,
+                    optimizer=self,
+                    optim_summary=self.optim_iter_summary,
+                    optimization_stats=self.system_simulator.stats
+                )
+            except WallTimeExceeded:
+                optim_result = self.write_state_to_result()
+
+        if self.system_simulator.stats is not None:
+            self.system_simulator.stats.end_t_opt = time.time()
 
         return optim_result
 
@@ -437,6 +720,9 @@ class SimulatedAnnealing(Optimizer):
     """
     This class uses simulated annealing for discrete optimization.
 
+    The use of this class requires the installation of the simanneal package
+    from pypi.
+
     Parameters
     ----------
     initial_temperature: float
@@ -449,8 +735,8 @@ class SimulatedAnnealing(Optimizer):
         Number of optimization iterations before the step size is reduced.
 
     bounds: array of boundaries, shape: (2, num_t, num_ctrl)
-        The boundary conditions for the pulse optimizations. bounds[0] should be
-        the lower bounds, and bounds[1] the upper ones.
+        The boundary conditions for the pulse optimizations. bounds[0] should
+        be the lower bounds, and bounds[1] the upper ones.
 
     """
 
@@ -459,6 +745,7 @@ class SimulatedAnnealing(Optimizer):
             system_simulator: Optional[simulator.Simulator] = None,
             termination_cond: Optional[Dict] = None,
             save_intermediary_steps: bool = False,
+            store_optimizer: bool = False,
             initial_temperature: float = 1.,
             final_temperature: float = 1e-6,
             step_size: int = 1,
@@ -466,16 +753,25 @@ class SimulatedAnnealing(Optimizer):
             bounds: Optional[np.ndarray] = None,
             updates: Optional[int] = None
     ):
+
+        try:
+            import simanneal
+        except ImportError as err:
+            raise RuntimeError(
+                'Requirements not fulfilled. Please install simanneal'
+            ) from err
+
         super().__init__(
             system_simulator=system_simulator,
             termination_cond=termination_cond,
-            save_intermediary_steps=save_intermediary_steps
+            save_intermediary_steps=save_intermediary_steps,
+            store_optimizer=store_optimizer
         )
 
         self.annealer = PulseAnnealer(
             state=0,
             bounds=bounds,
-            energy_function=self.cost_fktn_wrapper,
+            energy_function=self.cost_func_wrapper,
             step_size=step_size,
             step_ratio=step_ratio,
             Tmax=initial_temperature,
@@ -495,11 +791,16 @@ class SimulatedAnnealing(Optimizer):
         if self.system_simulator.stats is not None:
             self.system_simulator.stats.end_t_opt = time.time()
 
+        if self.store_optimizer:
+            storage_opt = self
+        else:
+            storage_opt = None
+
         optim_result = optimization_data.OptimizationResult(
             final_cost=costs,
             indices=self.system_simulator.cost_indices,
             final_parameters=pulse,
-            optimizer=self,
+            optimizer=storage_opt,
             optim_summary=self.optim_iter_summary,
             optimization_stats=self.system_simulator.stats
         )
@@ -538,6 +839,7 @@ class SimulatedAnnealingScipy(Optimizer):
             system_simulator: Optional[simulator.Simulator] = None,
             termination_cond: Optional[Dict] = None,
             save_intermediary_steps: bool = False,
+            store_optimizer: bool = False,
             temperature: float = 1.,
             step_size: int = 1,
             interval: int = 50,
@@ -546,22 +848,29 @@ class SimulatedAnnealingScipy(Optimizer):
         super().__init__(
             system_simulator=system_simulator,
             termination_cond=termination_cond,
-            save_intermediary_steps=save_intermediary_steps
+            save_intermediary_steps=save_intermediary_steps,
+            store_optimizer=store_optimizer
         )
         self.temperature = temperature
         self.step_size = step_size
         self.interval = interval
         self.bounds = bounds
 
-    def run_optimization(self, initial_control_amplitudes: np.ndarray):
+    def run_optimization(self, initial_control_amplitudes: np.ndarray,
+                         verbose: bool = False):
         """See base class. """
 
         super().prepare_optimization(
             initial_optimization_parameters=initial_control_amplitudes)
 
+        if self.store_optimizer:
+            storage_opt = self
+        else:
+            storage_opt = None
+
         try:
             result = scipy.optimize.basinhopping(
-                func=self.cost_fktn_wrapper,
+                func=self.cost_func_wrapper,
                 x0=initial_control_amplitudes.T.flatten(),
                 niter=self.termination_conditions["max_iterations"],
                 T=self.temperature,
@@ -569,7 +878,7 @@ class SimulatedAnnealingScipy(Optimizer):
                 take_step=self._take_step,
                 callback=None,
                 interval=self.interval,
-                disp=True
+                disp=verbose
             )
 
             if self.system_simulator.stats is not None:
@@ -582,7 +891,7 @@ class SimulatedAnnealingScipy(Optimizer):
                 num_iter=result.nfev,
                 termination_reason=result.message,
                 status=result.status,
-                optimizer=self,
+                optimizer=storage_opt,
                 optim_summary=self.optim_iter_summary,
                 optimization_stats=self.system_simulator.stats
             )
@@ -598,7 +907,7 @@ class SimulatedAnnealingScipy(Optimizer):
                 num_iter=self._n_cost_fkt_eval,
                 termination_reason='Maximum Wall Time Exceeded',
                 status=5,
-                optimizer=self,
+                optimizer=storage_opt,
                 optim_summary=self.optim_iter_summary,
                 optimization_stats=self.system_simulator.stats
             )
